@@ -47,9 +47,8 @@ class RemoteClosedException(Exception):
 
 class MessageReturnedException(Exception):
     def __repr__(self):
-        return 'Message %s was returned by RabbitMQ: (%s) %s' % (self.args[0],
-                                                                 self.args[1],
-                                                                 self.args[2])
+        return 'Message %s was returned by RabbitMQ: (%s) %s' % \
+               (self.args[0], self.args[1], self.args[2])
 
 class NoActiveTransactionError(Exception):
     def __repr__(self):
@@ -61,7 +60,6 @@ class TooManyChannelsError(Exception):
         return 'The maximum amount of negotiated channels has been reached'
 
 
-
 class UnexpectedResponseError(Exception):
     def __repr__(self):
         return 'Received an expected response, expected %s, received %s' % \
